#!/usr/local/bin/perl
#
# OMICHAIN
#
# XMM Pipeline Processing System OMICHAIN script
#
# Author: Jon Rainger (jfr@mssl.ucl.ac.uk)
#
# Description: This script runs the tasks required for to produce
#              PPS products from Imaging Mode ODFs.
#
# $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
#
#
#    MODIFICATIONS:
#    12/06/2000 - AMK - Removed call to OMPSF as none of the pipleline tasks use it's output.
#
#    12/06/2000 - AMK - Redirected sreen output to a file omichain_log as user will lose the output during a complete ODF
#
#    06/09/2000 - AMK - Ammended code so that the new style house keeping files are processed correctly.
#
#    06/09/2000 - AMK - Ammended code so that product names conform to specification.
#
#    29/09/2000 - AMk - Changed flat field product file name to tmp_flat_field as this file should NOT be a product.
#
#    11/10/2000 - AMK - Removed the die calls and replace with SAS::fatal calls.
#
#    03/11/2000 - CB  - Modified so that images with "BLOCKED" or "UNKNOWN" filters are not processed.
#                       Added call to ps2pdf.
#    21/01/2000 - CB - Modified so that "ommosaic" runs at the end to produce a composite image for each filter.
#    10/08/2001 - CB - Modified to process engineering 2 and 4 data and restructured.
#    11/20/2001 - CB - Modified to accept command-line parameters for ommodmap, omdetect, ommodmap, ommosaic
#                      and omsrclistcomb.     
#    12/12/2001 - CB - Modified to process the exposures for each filter in the order that the filters 
#                      appear in the SAS summary file.                       
#    08/-5/2002 - CB - Modified to keep the intermediate source-list files from omdetect & ommag
#    03/12/2002 - CB - Modified to use the omatt parameter rotateimage
#************************************************************************************************************************
#use DAL;
#use SAS;


#sub omichain 
{
#/////////////////////////////////////////

    @FilterLst =();                   # Stores a list of the filters in the summary file
    @ExposureNoList =();               # Stores a list of the exposure numbers
    @ExposureFilterList =();           # Stores a list of the filter for each exposure number
    @E2IMIList =();                    # Stores the IMI files created by omengine2
    @E2IMIList =();
    @E4IMIList =();                    # Stores the engineering 4 mode image files
    @E4IMIList =();
    @Simage2000Files = ();   
    %ExposureType = ();
    $SASFile = "TEST";
    
    $omatt_rotateimage = T;



    #************************************************
    # For program development message can be written
    # to a file "omichain_log"
    # REMEMBER TO BLANK OUT NEXT LINE BEFORE UPLOADING
    #***********************************************
    
   
    &CreateLogFile;


    #**************************************
    # Set up the directory paths for input
    # and output.
    #**************************************



    $inp_directory = "";
    $out_directory = "";
    &SetUpDirectoryPaths;

     

    my $taskname='OMICHAIN';
      

    my $VERSION='1.33';
    my $date = &date_time;
#    $comment = stringParameter("comment");
     

    my $create = "Running SAS task $taskname V$VERSION $date";
      

    my $text1 = "Input directory = $inp_directory";  
    my $text2 = "Output directory = $out_directory"; 
      
 
    &HighlightedMessage("*", $create, $comment, "   ", $text1, $text2);
    
  
   
    #**************************************
    # Get the parameters for ommodmap,
    # omdetect, omatt, omsrclistcomb & ommosaic
    #**************************************
    &GetOmmodmapParameters;

    &GetOmdetectParameters;

    &GetOmattParameters;
    

    &GetOmsrclistcombParameters;

    &GetOmmosaicParameters;

    #**************************************
    # Get the list of filters that the
    # user wants to process (if any in list)
    #**************************************
      
  
    @InputFilterList = &GetStringParameterList("filters");
      
   
    
    #***************************************************
    # Initialise the lists of engineering 2/4 data, etc.
    #***************************************************
    &InitialiseLists;

     
   
  
    #************************************************
    # Get a list of the files in the current directory
    #************************************************
    &GetListOfDataFiles;
    
   
    #************************************************** 
    # Check for the presence of the house keeping files
    #**************************************************
    &CheckForHouseKeepingFiles;

    #**********************************************
    # Run the pipeline- Loop over all observations.
    #**********************************************
    foreach $orbitNo (keys(%orb_list))
    {


        #********************************************
        # Process the data for each orbit in the list
        #********************************************
        &ProcessOrbit($orbitNo, @InputFilterList);  

    }
  
    #********************************************
    # We don't want to keep any new image file
    # created from Engineering mode 2 data
    #********************************************
#    &DeleteNewImageFiles;

    $date = &date_time;
       
    $message = "Finished running SAS task $taskname V$VERSION $date";
    &HighlightedMessage("*", $message);

    

    #******************************
    # For test purposes
    #******************************
    if ($redirect == 5)
    {
        close(STDOUT);
        close(STDERR);
    }

}

#***********************************************************
#
# Subroutine section
#
#***********************************************************
sub date_time {          # create the file creation datestring

    local $datestring;
    $datestring = `date -u`;
    chop $datestring;
    return $datestring;
}
#**************************************
# subroutine GetFilterCode
# Given a filter from the SAS file
# returns the filter code to be used
# in constructing the output file name.
#**************************************
sub GetFilterCode
{
    my($filter) = @_;  # The input filter name
    
    
    @OMFilters = ("BLOCKED", "V",  "Magnifier", "MAGNIFIER", "U",  "B",  "WHITE", "GRISM2", "Grism2", "UVW1", "UVM2", "UVW2", "GRISM1", "Grism1");

    @OMFilterCodes = ("BL",  "V", "MA", "MA", "U", "B", "W", "G",  "G",  "L",   "M",   "S", "H", "H");


    for ($i = 0; $i < @OMFilters; $i++)
    {
       
        if ($OMFilters[$i] eq $filter)
        {
            return $OMFilterCodes[$i];
        
	}
    }
    die("No such filter ($filter) in filter table!");
    
}
#************************************
# Sub IsHighResCentralWindow
#************************************
sub IsHighResCentralWindow
{
    local ($fileName) = @_;
                 
    if($fileName =~ /OMS00/)
    {
       
        if($fileName =~ /IMAGE_1000/)
        {
            return 1;
       
        }     
    }    
    return 0;
}
#********************************************************************
# subroutine FindEngineering4ImageFiles
#********************************************************************
sub FindEngineering4ImageFiles
{
    $numExposures = @ExposureNoList;
  
    @ImageFile = <$inp_directory/*E4I.FIT*>;        

    $numFiles = @ImageFile;
   
    if($numFiles == 0) 
    {
        return;   
    }
    #************************************************************************
    # Create a list of engineering 4 image files, without the directory path
    #************************************************************************
    for ($i = 0; $i < $numFiles; $i++)
    {
        $fileName = $ImageFile[$i];       
        $pos = index($fileName, "E4I") - 24;
        $name = substr($fileName, $pos, 32);    # Get rid of the directory name                        
        push(@E4IMIList);                       # Add it to the list of engineering mode 4 data


        $fileName = $name;

        #*********************************************
        # Find out which exposure this file is
        #*********************************************
        $exposure = substr($fileName, 19, 3);      # The exposure number
    
        $found = 0;
        $j = 0;

        while ($j < $numExposures and $found == 0)
        {
            
            $exposureNo = substr($ExposureNoList[$j], 0, 3);
        
            if ($exposure eq  $exposureNo)
            { 
                $ExposureType{$exposureNo} = "ENG4";
                $ExposureExtension{$exposureNo} = "*E4I.FIT*";
	    
		$found = 1;
                       
            }
            $j++;
        }

    }
}
#******************************************************
# Subroutine FillExposureList
# Obtains a list of the exposures in the summary file
#******************************************************  
sub FillExposureNoList
{


    my($SASFILE) = "SASFILE";
 
  

   open ($SASFILE, $SASFile) || die("Unable to open SAS SUMMARY FILE $SASFILE"); 

   
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************
 

    $line = <$SASFILE>;

       
    until (substr($line, 0, 2) eq "OM")
    {
       $line = <$SASFILE>;
                                  
    } 

   
    while ($line = <$SASFILE>)

    {
       
        if ($line =~ "/ Exposure Identifier")
        {        
 
	  
            $exposureNo = substr($line, 0, 3);
            $exposureSchedule = substr($line, 32, 1); 
            push(@ExposureNoList, $exposureNo);      
                 
            $ExposureScheduleList{$exposureNo} = $exposureSchedule;
            #****************************************
            # Get the filter name from the SAS file
            #****************************************
	  
            $filterName = &GetNextSASFilter($SASFILE);
	   
            if ($filterName eq "")
            {
                last;
            }
            $ExposureFilterList{$exposureNo} = $filterName;

	    
	       
        }
    }
#    &Display("LIST OF EXPOSURES", @ExposureNoList);
#    &Display("Exposure Filter List", @ExposureFilterList);

#    &Display("Exposure Filter List", @ExposureFilterList);
    
    close $SASFILE;
    
 

}
#***************************************************************
# Subroutine DeleteNewImageFiles
# Renames the new image files created from each set of four
# engineering 2 mode image files.
#***************************************************************
sub DeleteNewImageFiles
{

    if ($out_directory ne $inp_directory)
    { 
        $num = @E2IMIList;   


        for ($j = 0; $j < $num; $j++)
        {
        
            $fileName = $E2IMIList[$j];       
            $pos = index($fileName, "E2I") - 24;
            $name = $out_directory . '/' . substr($fileName, $pos, 32);    # Get rid of the directory name               
    
      
            system("cp $fileName $name"); 
        }         
        	    		
        unlink (@E2IMIList);
    }
              


    #*********************************************
    # Move the flatfield file created by omflatgen
    # to the output directory and delete the one
    # in the input directory
    #*********************************************
    if ($flatFieldFileName ne "BLANK" and $out_directory ne $inp_directory) 
    {
        $name = $inp_directory . '/'. $flatFieldFileName;
        system("cp $flatFieldFileName  $name"); 
        unlink $flatFieldFileName;
    }

    #******************************************
    # Delete temp files
    #*****************************************
    chdir $out_directory;

    &DeleteFileList("tmp");

    #******************************************
    # Delete imi files
    #*****************************************
    &DeleteFileList("imi");
 
    #******************************************
    # Delete fit files
    #*****************************************
    &DeleteFileList("fit");
 

}
#***************************************
# DeleteFileList
#***************************************
sub DeleteFileList($)
{ 
    my ($name) = @_;
   
    @List = <*$name*> ;    # Creates a list of FIT file names    

    unlink (@List);
}

#**************************************************************
# Subroutine MissingImageFileWarning
#*************************************************************
sub MissingImageFileWarning($)
{ 


#    my ($fileName) = @_;
    $exp_key = " ";
    $value = $ExposureType{$exp_key};
    
    if ($value eq "ENG2" or $value eq "ENG4" and $value ne "")
    {
        return;
    }

#    $im_file = $orb_key . "_" . $obs_key . "_OM" . $exp_type . $exp_key . $win_key . "IMI.FIT";
    
	
    &Message("*** Warning *************************************************");

#    $message = "Warning! - Window '$win_key' of exposure '$exp_key' in observation '$obs_key' is missing the image file $im_file.";
    &HighlightedMessage("*", $message);
   
  
    
}
#********************************************
# Subroutine CreatePoductFileNames
#********************************************
sub CreateProductFileNames($)
{
    
    my ($observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter) = @_;           # The exposure and window numbers
    $windowNo = substr($windowNo, 1, 1);
    $exposureType = $ExposureType{$exposureNo}; # Rudi5, ENG2 or ENG4

     
    my $filterCode = &GetFilterCode($filter);
     
    if ($exposureType eq "RUDI5")       # Rudi-5 case
    {
        $string1 = "SIMAGE$windowNo";
        $string2 = "IMAGE_$windowNo";
        $string3 = $windowNo;
    }
    elsif ($exposureType eq "ENG2")     # Engineering-2 case
    {
        $string1 = "FSIMAG$filterCode";
        $string2 = "FIMAG_$filterCode";
        $string3 = $filterCode;
 
    }
    elsif ($exposureType eq "ENG4")     # Engineering-4 case
    {
        
        $string1 = "FSIMAG$filterCode";
        $string2 = "FIMAG_$filterCode";
        $string3 = $filterCode;
 
    }
    else
    {
        die("UNKNOWN EXPOSURE TYPE ($exposureType)");
    } 
    
    
    #***************************************************
    # The detector-coordinate (ie non-rotated image) file name
    #***************************************************
    $detectorCoordImageFileName = $out_directory . "/P$observationNo" . "OM$exposureSchedule$exposureNo$string2" . "000.FIT";
    

    #***************************************************
    # The sky-coordinate (ie rotated image) file name
    #***************************************************
    $skyCoordImageFileName = $out_directory . "/P$observationNo" . "OM$exposureSchedule$exposureNo$string1" . "000.FIT";

    

    #************************************
    # The intermediate image file names
    #***********************************
    $IntermediateImage1FileName = $out_directory . "/I$observationNo". "OM$exposureSchedule$exposureNo" . "IMAGE1$windowNo" . "000.FIT";
    $IntermediateImage2FileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "IMAGE2$windowNo" . "000.FIT";
#    $IntermediateImage3FileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "IMAGE3$windowNo" . "000.FIT";
    $flat_field_product_filename = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "FLATF$windowNo" . "000.FIT";

    #******************************
    # Product created by ommodmap
    #*****************************
    $modulo_8_product_filename = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "MOD8$windowNo" . "000.FIT";

    #*****************************
    # Products created by omdetect
    #*****************************
    $regionFileName = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "REGION$windowNo" . "000.ASC";

    $levelImageFileName = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "LEVELIMAGE$windowNo" . "000.FIT";



    #**************************************************
    # The source=list created by omdetect
    #**************************************************
    $osw_list_intermediary_detect_filename1 = "I$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI1$string3"."000.FIT";


    #**************************************************
    # The source-list created by ommag
    #**************************************************
    $osw_list_intermediary_detect_filename2 = "I$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI1$string3"."001.FIT";
    
    #**************************************************
    # The source-list created by ommat (the product file)
    #**************************************************
   
    
    $osw_list_intermediary_detect_filename = "P$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI$string3"."000.FIT";



}
#********************************************
# Subroutine CheckTHXFileExits
#********************************************
sub CheckTHXFileExists($)
{
    
    my ($orbitNo, $observationNo, $exposureSchedule, $exposureNo) = @_;           # The exposure and window numbers
 
   
 
    if (exists(${$orbitNo . $observationNo . $exposureSchedule . $exposureNo . "00"}{THX})) 
    {
 

	$tracking_history_filename = $inp_directory . "/". ${${$orbitNo . $observationNo . $exposureSchedule . $exposureNo . "00"}{"THX"}};
        


    }	 
    else    
    {
    
        #***********************************
        # Create a dummy name to fool OMPREP
        #***********************************
        $tracking_history_filename = "DUMMYTHX.FIT";

 
    }

}

#************************************************
# Subroutine ProcessIageFile
# Processes an OM imaging file
#***********************************************
sub ProcessImageFile($)
{ 

    my ($image_filename, $observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter) = @_;   

 
    $exposureType = $ExposureType{$exposureNo};   # Rudi5, ENG2 or ENG4
   

    #******************************************
    # Set pipeline variables.
    #**************************
    &CreateProductFileNames($observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter);


    #***********************************
    # Produce image mode products.
    #***********************************

    &HighlightedMessage("*", "Processing image file $image_filename");
    #*********************************
    # omprep
    #*********************************
    &HighlightedMessage("*", "omprep");
  

    $arg_list = "set=$image_filename" .
                " nphset=$inp_directory/$non_periodic_hk_filename" .
                " pehset=$inp_directory/$periodic_hk_filename" .
                " wdxset=$inp_directory/$window_data_filename" .
		" outset=$IntermediateImage1FileName";

    &Message("omprep $arg_list");

    #**************************************************************
    # Modified so that if omprep fails it just skips the observation,
    # but does not kill the chain (1.11.2000)
    #**************************************************************
    system("omprep $arg_list");

    $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value == 1) # Failure in omprep 
    { 
        &HighlightedMessage("*", "omprep has detected an error- this observation will not be processed");
        return;

    }     
    
    #******************************************
    # omcosflag
    #******************************************
    &HighlightedMessage("*", "omcosflag");

    $arg_list = "set=" .  $IntermediateImage1FileName .
                " thxset=" . $trackingHistoryFileName;

    &Message("omcosflag $arg_list");
    system("omcosflag $arg_list"); #  && SAS::fatal("Task Failure","omcosflag failed.");
 
  
    #****************************************
    # omflatfield
    #****************************************
    &HighlightedMessage("*", "omflatfield");

    $arg_list = "set=$IntermediateImage1FileName" .
                " thxset=$trackingHistoryFileName" .
                " inorbitflatset=$in_orbit_flat_field_filename" .
                " tsflatset=$flat_field_product_filename" .
                " outset=$IntermediateImage2FileName";

    &Message("omflatfield $arg_list");

    system("omflatfield $arg_list"); #  && SAS::fatal("Task Failure","omflatfield failed.");
               


    #***************************************
    # ommodmap
    #***************************************
    &HighlightedMessage("*", "ommodmap");

    $arg_list = "set=$IntermediateImage2FileName" .
                " mod8product=yes" .
		" mod8set=$modulo_8_product_filename" .
                " outset=$detectorCoordImageFileName"; 
#                " outset=$IntermediateImage3FileName" .
#                " nsig=$ommodmap_nsig" .
#                " nbox=$ommodmap_nbox";  

    &Message("ommodmap $arg_list");
    system("ommodmap $arg_list"); # && SAS::fatal("Task Failure","ommodmap failed.");
                     


    #***********************************************
    # Stop processing image file if a grism image
    #***********************************************
  
   if ($filter eq "GRISM1" or $filter eq "GRISM2" or $filter eq "MAGNIFIER" or $filter eq "Magnifier") 
    {
        &HighlightedMessage("X", "Cannot process this image file any further ($filter image)");
        return 1;


    }
    #**************************************************
    # omdetect
    #**************************************************
    &HighlightedMessage("*", "omdetect");
    $arg_list = #"nsigma=$omdetect_nsigma" .
#                " minsignificance=$omdetect_minsignificance" .
#                " detectextended=$omdetect_detectextended" .
                " set=$detectorCoordImageFileName" .
                " regionfile=$regionFileName" .
                " levelimage=$levelImageFileName" .
                " detectextended=f" .   
                " outset=$out_directory/$osw_list_intermediary_detect_filename";

 
    &Message("omdetect $arg_list");

    system("omdetect $arg_list"); #  && SAS::fatal("Task Failure","omdetect failed.");

    #*********************************************************************
    # Keep a copy of this source-list
    #*********************************************************************

    system("cp $out_directory/$osw_list_intermediary_detect_filename $out_directory/$osw_list_intermediary_detect_filename1");
    #*********************************************** 
    # ommag
    #***********************************************
    &HighlightedMessage("*", "ommag");

    $arg_list = "set=$out_directory/$osw_list_intermediary_detect_filename";
                

    &Message("ommag $arg_list");

    system("ommag $arg_list"); #  && SAS::fatal("Task Failure","ommag failed.");

    #*********************************************************************
    # Keep a copy of this source-list
    #*********************************************************************
    system("cp $out_directory/$osw_list_intermediary_detect_filename $out_directory/$osw_list_intermediary_detect_filename2");


    #***********************************************
    # omatt
    #***********************************************
    &HighlightedMessage("*", "omatt");

    
    $arg_list ="set=$detectorCoordImageFileName" .
        
#" set=$IntermediateImage3FileName" .
           #"set=$detectorCoordImageFileName" .
                " sourcelistset=$out_directory/$osw_list_intermediary_detect_filename" . 
		    " ppsoswset=$skyCoordImageFileName usecat=t" .
                    " catfile=$inp_directory". "/usnocat.fit";
#                " usecat=$omatt_usecat" .
#		" rotateimage=$omatt_rotateimage";
  
   
   


    &Message("omatt $arg_list");
    system("omatt $arg_list"); #  && SAS::fatal("Task Failure","omatt failed.");
    
    #**********************************************************
    # 12.6.2001
    # Add the file to the list if it is not the high-resolution
    # central image
    #**********************************************************
                  

#  system("cp $out_directory/$osw_list_intermediary_detect_filename  $out_directory/eposcorr.fit");
#    $arg_list ="xrayset=$out_directory/eposcorr.fit". 
#	       " opticalset=$inp_directory". "/usnocat.fit opticaltableext=SRCRES";


  

 
#    &Message("eposcorr $arg_list");
#    system("eposcorr $arg_list");
    



    if (&IsHighResCentralWindow($osw_list_intermediary_detect_filename) == 0)
    {                   
	$name = $out_directory . '/' . $osw_list_intermediary_detect_filename;
        push(@osw_file_list, $out_directory . '/' . $osw_list_intermediary_detect_filename);
	if ($skyCoordImageFileName =~ "SIMAGE00")
        {
            return;           
	}
        print("Line 797 pushing     $skyCoordImageFileName onto SimnageList\n");
        push(@Simage2000Files, $skyCoordImageFileName);
              
 
    }
       

}
#************************************************************
# Subroutine GetListOfDataFiles
# Get a list of file names in the current working directory.
#************************************************************
sub GetListOfDataFiles
{
    opendir(INDIR, $inp_directory); #  || SAS::fatal("OMICHAIN","Unable to open $inp_directory"); 

    $periodic_file_flag = 0;
    $non_periodic_file_flag = 0;

    #******************************************************************************
    # Sort the file names according to observation, exposure, window and file type.
    #******************************************************************************

    foreach $file (glob("$inp_directory/*"))

    {
        $file =~ s/^.+\///;
        print("file=$file\n");
        if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})([0-9A-Z]{3}).FIT/i)
  
        {
            if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM[S|U|X][0-9a-zA-Z]{3}[0-9]{2}FFX.FIT/i) 
            {
                $ffx_list{"$1$2"} = "$inp_directory/$file";
            }
            else
            {

#***************************************************************************************************
# We need to filter out the housekeeping files for this loop. There is now only one PEH and NPH file 
# per ODF.
#
#
# We are now picking up the exposure type flag and this stored with the exposure number.
# This is necessary as we are required to insert the exposure type flag into the product file names. 
# This necessitates some further filtering below.
#****************************************************************************************************
 
  
	        if ($file !~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})PEH.FIT/i
	           && $file !~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})NPH.FIT/i)
	        {
                    #******************
                    # Orbit number list
                    #******************
	            $orb_list{$1} = "$1";
                    #****************** 
                    # Observation list.
                    #******************
	            ${${"$1"}{$2}} = "$1$2";
                    #**********************************************
                    # Exposure list.
                    # This contains a value like 'X123'. X is the exposure flag and 123 i the exposure number.
                    #**********************************************
	            ${${"$1$2"}{$3}} = "$1$2$3";
                    #*************
                    # Window list.
                    #************* 
	            ${${"$1$2$3"}{$4}} = "$1$2$3$4";
                    #************************************************
                    # File list.
                    # Check for E2I.FIT & E4I.FIT file
                    # and make sure it gets stored in the IMI list
                    #************************************************
       
                    if ($5 eq "E2I" or $5 eq "E4I")
                    {
                        $num = "IMI";
                        ${${"$1$2$3$4"}{$num}} = "$file";



                    }
                    else
                    {
                        ${${"$1$2$3$4"}{$5}} = "$file";
                    }

                 

                }
                else
                {
                #*****************************************
                # Pick-up the House Keeping files here.
                #*****************************************
 
	            if ($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})PEH.FIT/i)
	            {
		        $periodic_hk_filename = $file;

		        $periodic_file_flag = 1;
	            }

	            if($file =~ /^([0-9A-Z]{4})_([0-9A-Z]{10})_OM([[0-9a-zA-Z]{4})([0-9]{2})NPH.FIT/i)
	            {
		        $non_periodic_hk_filename = $file;

		        $non_periodic_file_flag = 1;
	            }
                }
            }
        }
    }
    closedir(INDIR);

}
#*********************************************
# GetFlatFieldImage
#*********************************************
sub GetFlatFieldImage($)
{
    my ($orb_key, $obs_key, $exp_type) = @_;   
  

    if (exists($ffx_list{$orb_key . $obs_key}))
    {
        $in_orbit_flat_field_filename =  $ffx_list{$orb_key . $obs_key };
        $flatFieldFileName = "BLANK";
    
    }   
    else
    { 
        &HighlightedMessage("*", "omflatgen");
      
        $flatFieldFileName = "P$orb_key" . $obs_key . "OM$exp_type" . "000FLAFLD0000.FIT";
        $ffx_list{"$1$2"} = $flatFieldFileName;
        $flatFieldFileName = $out_directory . "/$flatFieldFileName";
        $arg_list = "outset=$flatFieldFileName";

        $message = "omflatgen $arg_list";
        &Message($message);
        #********************************
        # Create an artificial one
        #********************************

        system("omflatgen $arg_list"); #  && SAS::fatal("Task Failure", "omflatgen");

         $in_orbit_flat_field_filename = $flatFieldFileName;  
      
    }
  
    return ($in_orbit_flat_field_filename);

}
#****************************************************
# Subroutine ProcessExposure
#****************************************************
sub ProcessExposure($)
{
    my ($orbitNo, $observationNo, $exposureNo, $exposureSchedule, $filter) = @_; 
    my ($type) = $ExposureType{$exposureNo};     # One of RUDI5, ENG2 or ENG4
 
    #***********************************************
    # Determine if RUDI5, ENG2 or ENG4 exposure
    #***********************************************

    $exposureType = &FindExposureType($exposureNo);
   

    if ($exposureType eq "MISSING")
    {
     
        &HighlightedMessage("X", "Warning! - No image files found for exposure $exposureNo - ignoring exposure");
        return;    
          
    }
    $text = "Processing exposure no $exposureNo ($ExposureType{$exposureNo})";
    &HighlightedMessage("*", $text);


    my ($exposureKey) = $exposureType . $exposureNo;
    my ($i) = 0;


    #**************************************************
    # If the exposure is an engineering-2 one combine
    # the four separate images into one.
    #**************************************************
    if ($exposureType eq "ENG2")
    {
        &CombineEng2Images($observationNo,  $exposureSchedule, $exposureNo);
        $directory = $out_directory;
    }
    else
    {
        $directory = $inp_directory;
    }

   
    #*******************************************************************
    # Check consistancy of odf.
    #
    # We need only check for the presence of the WDX and THX files here.
    #
    # We now have a true or dummy tracking history file name
    #*******************************************************************

    &CheckTHXFileExists($orbitNo, $observationNo, $exposureSchedule, $exposureNo);
   
    $window_data_filename = $orbitNo . "_" . $observationNo . "_OM" . $exposureSchedule . $exposureNo . "00WDX.FIT";
    if (&FileExists($window_data_filename == 0))
    {    
        $text = "Missing WDX file- skipping exposure";
        &HighlightedMessage("X", $text);
        return;     
    }     

    #**************************
    # Set pipeline variables.
    #**************************
  
    $tracking_history_plot_filename = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSHPLT" . "000.ps";

     
    $tracking_history_plot_filename1 = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSHPLT" . "000.PDF";

    $tracking_history_timeseries_filename = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSTRTS000.FIT";

    #************************************
    # Tracking pipeline
    #************************************
    &HighlightedMessage("*", "Running Tracking pipeline");

    #********************************************
    #  omprep
    #********************************************
    &HighlightedMessage("*", "omprep");

    #********************************
    # The tracking-history file name
    #*******************************
    $trackingHistoryFileName = $out_directory . "/I" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TRHIS0000.FIT";


    $arg_list = "set=$tracking_history_filename" .
                " nphset=$inp_directory/$non_periodic_hk_filename" .
                " pehset=$inp_directory/$periodic_hk_filename" .
                " wdxset=$inp_directory/$window_data_filename" .
                " outset=$trackingHistoryFileName" .
                " modeset=3";                          

    system("omprep -v");
    &Message("... omprep  $arg_list"); 
 
    system("omprep  $arg_list");

    $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value == 1) # Failure in omprep 
    { 
        &Message("omprep has detected an error- this observation will not be processed");
        return;  
    }     
   

    #************************************************************
    # omdrifthist
    #************************************************************
    &HighlightedMessage("*", "omdrifthist");

    $arg_list = "set=$trackingHistoryFileName" .
                " plotfile=" .  $tracking_history_plot_filename;

    &Message("omdrifthist $arg_list");

    system("omdrifthist $arg_list"); #  && SAS::fatal("Task Failue","omdrifthist failed.");
	     

    #*******************************************************************
    # Only run ps2pdf if tracking-history file exists
    #*******************************************************************
    
      
    if (&FileExists($tracking_history_plot_filename)) 
    {

        &HighlightedMessage("*", "ps2pdf");

        $arg_list = "$tracking_history_plot_filename $tracking_history_plot_filename1";
        &Message("ps2pdf  $arg_list");
             
	  
        system("ps2pdf $arg_list"); #  && SAS::fatal("Task Failure","ps2pdf failed.");
     
    }
    else
    {
        &HighlightedMessage("X", "Tracking-history-plot file name doesn't exist");
      
    }
    #************************************************************
    # omthconv
    #************************************************************
    &HighlightedMessage("*", "omthconv");

    $arg_list = "thxset=$trackingHistoryFileName" .
                " nphset=$inp_directory/$non_periodic_hk_filename" .
                " outset=$tracking_history_timeseries_filename" ;

    &Message("omthconv $arg_list");

    system("omthconv $arg_list"); #  && SAS::fatal("Task Failure","omthconv failed.");
    

    #***************************************************
    # Loop over the windows (0 or 1) for RUDI5 images
    # (Low res & high res window)
    #***************************************************


    #**************************************************
    # Get a list of the IMI/E2I/E4I for this exposure
    #**************************************************
    my $extension = $ExposureExtension{$exposureNo};     # One of RUDI5, ENG2 or ENG4

    #***********************************************
    # Get a list of image files for this exposure
    #***********************************************

    my @FileList = &GetExposureFileList($directory, $exposureNo, $extension);

    #************************************************
    # If the exposure is a Rudi-5 one check that
    # there are 2 exposures (high and low resolution)
    #************************************************

    if ($exposureType eq "RUDI5")
    {
        &CheckRudi5Images($exposureNo, @FileList);
    } 

    #*********************************************
    # The imaging pipeline
    #*********************************************

    &HighlightedMessage("*", "Running Imaging pipeline");



    for ($i = 0; $i < @FileList; $i++)
    {
                                     
        my $imageFileName = $FileList[$i];

         
        $pos = index($imageFileName, $extension) - 24;
        $name = substr($imageFileName, $pos, 32);    # Get rid of the directory name               
        $windowNo = substr($name, 16, 2);            # Window no- 00 or 01
               
        &ProcessImageFile($imageFileName, $observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter);
  
    }

}
#***************************
# Subroutine ProcessOrbit
#***************************
sub ProcessOrbit($)
{
    my ($orbitNo, @UserFilterList) = @_; 
   
       
    $obs_list = $orb_list{$orbitNo};
   
    foreach $obs_key (keys(%{$obs_list}))
    {
       

        &HighlightedMessage("*", "Observation $obs_key");
        
    
        #**************************************************
        # Process the data for each exposure of each filter
        #**************************************************
  
        @UserExposureList = &GetStringParameterList("exposures");


        if (@UserFilterList > 0 and @UserExposureList > 0)
        {
            &HighlightedMessage("X", "Error! - You can only supply either a list of filters or a list of exposures- aborted");
            die;    
        } 
        my $userSupplied = 0;
        my @List = 0;
        if (@UserFilterList > 0)
        {
            #*****************************************************
            # Check that the list is valid and modify accordingly
            #*****************************************************
            @List = &CheckUserFilterList(@UserFilterList);
             
        }
        elsif (@UserExposureList > 0)
        {
           

            #*****************************************************
            # Obtain the list of filters for the user-specified
            # exposures
            #*****************************************************
            @List = &CheckUserExposureList(@UserExposureList);



            $userSupplied = 1;
        }
        else
        {
            @List = &GetFilterList;  # List of OM filters in summary file
        }


        #*************************************************
        # Get the flat-field (create one if necessary)
        #*************************************************
#        $exp_type = " ";     
        $in_orbit_flat_field_filename = &GetFlatFieldImage($orbitNo, $obs_key, $exp_type); 
       


        #**************************************
        # Get the list of filters that the
        # user wants to process 
        #**************************************

         
  
        my $k = 0;
 
      
        for ($k = 0; $k < @List; $k++)
        {
	  
	 
            @Simage2000Files = () ;    # Stores a list of the image files for mosaicing    


            my @List1 = &GetExposureNoListForFilter($List[$k]);
            if ($userSupplied == 1)
            {
		
                my @temp = ();
                my $l = 0;
                for ($l = 0; $l < @UserExposureList; $l++)
                {
                    my $m = 0;
                    my $found = 0;
                    while ($m < @List1 and $found == 0)
                    {
                        if($UserExposureList[$l] eq $List1[$m])
                        {
                            $found = 1;
                            push(@temp, $UserExposureList[$l]);
                        }
                        $m++;
                    }
                 }
                 @List1 = @temp;

 

      
	     } 

            my $noExposures = @List1;
	    if ($noExposures == 1)
            {
                $message = "Processing $List[$k] band image (exposure " . join(", ", @List1) . ")";   
            }
            else
            {    
                $message = "Processing $List[$k] band images (exposures " . join(", ", @List1) . ")";   
            }
	    &HighlightedMessage("*", $message);
             
                    
            my $start = time();
            my $i = 0;

            for ($i = 0; $i < $noExposures; $i++) 
            {

                my $exposureNo = $List1[$i];
	       
                if ($ExposureType{$exposureNo} ne "MISSING")
                {
                    $exposureType = $ExposureType{$exposureNo};
                } 
                my $exposureSchedule = $ExposureScheduleList{$exposureNo};
         
                my $exposure_key = $exposureSchedule . $exposureNo;
         
           
                &ProcessExposure($orbitNo, $obs_key, $exposureNo, $exposureSchedule,  $List[$k]);
	       

            }

            #****************************************
            # Mosaic the images for thsi filter
            #****************************************    
 	    my $num =  @Simage2000Files;
	    print("Line 1310 num of simagefiles=$num\n");
            if ($num > 1)
            {
                 
                &MosaicImages($exposureType, $List[$k], @Simage2000Files);    

            }
            else
            {
                if ($num == 1)
                {
                    $zero = 0;
                    &HighlightedMessage("*", "Warning- there is only one sky-coord image ($Simage2000Files[$zero]" . "- ommosaic will not run");
                }

            } 
            my $time = time() - $start;
            if ($time > 1)
            {
                $message = "Finished processing images for filter $List[$k] - time taken= $time  seconds";
	        &HighlightedMessage("*", $message);
            }       
        } 

        #***********************************************
        # Check that there is at least one source-list
        #***********************************************
        my $numSourceLists = @osw_file_list;
        if ($numSourceLists > 0) 
        {  


            #**********************************
            # Set remaining pipeline variables.
            #**********************************
 
#********************  what is exp_type ***************************

            $observation_source_list_pps_product_filename = "P$obs_key" ."OM$exposureType" . "000" . "OBSMLI0000.FIT";

            #**********************************************
            # omsrclistcomb.
            #********************************************** 
            &HighlightedMessage("*", "omsrclistcomb");

            $arg_list = "sourcelistsets=\"" . join(" ", @osw_file_list) . "\"" .   
#		        " nsigma=$omsrclistcomb_nsigma" .
#                        " usecat=$omsrclistcomb_usecat" .
                        " usecat=t" .
                        " catfile=$inp_directory". "/usnocat.fit" .
                        " outset=$out_directory/$observation_source_list_pps_product_filename";

            &Message("omsrclistcomb $arg_list");

            system("omsrclistcomb $arg_list"); #  && SAS::fatal("Task Failure","omsrclistcomb failed.");


#    system("cp $out_directory/$observation_source_list_pps_product_filename $out_directory/eposcorr.fit");
#    $arg_list ="xrayset=$out_directory/eposcorr.fit". 
#	       " opticalset=$inp_directory". "/usnocat.fit opticaltableext=SRCRES";


  

 



#    &Message("eposcorr $arg_list");
#    system("eposcorr $arg_list");
    



        
            #************************************************
            # ommosaic
            #************************************************          
#            &RunOmmosaic;
        }
        else
        {
            &HighlightedMessage("Warning!- There are no source-list files for omsrclistcomb to process");
        }
    }

}
#*******************************
# sub SetUpDirectoryPaths
#*******************************
sub SetUpDirectoryPaths
{
  
#    $inp_directory = stringParameter("inpdirectory");
    if ($inp_directory eq "")
    {


        #*************************************************
        # If the SAS_ODF environment variable has been set
        # then set the input directory to it- otherwise
        # set the input directory to the current directory
        #*************************************************
        $input = $ENV{SAS_ODF};
       
        if ($input eq "")
        {
            $inp_directory = &GetCurrentDirectory;
            $ENV{SAS_ODF} = $inp_directory;
            @SASFiles = <$inp_directory/*.SAS>;        
            $SASFile = shift @SASFiles;
   
           
        }
        else
        {
	   
            #*************************************************
            # If SAS_ODF is set to a .SAS file then need
            # to get the ODF directory from it
            #************************************************

            $pos = index($input, "SAS");
            if ($pos > 0) 
            {
	        $length = length($input);
                $directory = substr($input, 1, $length - 4); 
                $SASFile = $input;

		$inp_directory = &GetODFDirectory($input);
	       
            }
            else
            {
	      

                $inp_directory = $input;

                @SASFiles = <$inp_directory/*.SAS>;        

 
                $SASFile = shift @SASFiles;
      
	    
   
            }
        }

    }
   
#    $out_directory = stringParameter("outdirectory");


    if ($out_directory eq "")
    {
	$out_directory = &GetCurrentDirectory;
    }
   
 
}
#**********************************
# sub CheckForHouseKeepingFiles
#**********************************
sub CheckForHouseKeepingFiles
{
    if($periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Periodic Housekeeping file. Please insert into the ODF. ");
#        SAS::fatal("Task Failure","Incomplete ODF.");
    }

    if($non_periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Non Periodic Housekeeping file. Please insert into the ODF.");
#        SAS::fatal("Task Failure","incomplete ODF.");
    }
}

#**************************************
# Subroutine InitialiseLists
#**************************************
sub InitialiseLists
{
#    @FilterList =();                   # Stores a list of the filters in the summary file
#    @ExposureNoList =();               # Stores a list of the exposure numbers
#    @ExposureFilterList =();           # Stores a list of the filter for each exposure number
#    @E2IMIList =();                    # Stores the IMI files created by omengine2
#    @E2IMIList =();
#    @E4IMIList =();                    # Stores the engineering 4 mode image files
#    @E4IMIList =();

#    %ExposureType = ();

    &FillExposureNoList;           # Stores a list of exposures from the summary file. 

   
}
#**************************************
# Subroutine Message
#**************************************
sub Message($)
{ 
    my ($message) = @_;  
    if ($redirect == 5) 
    {
        print("$message \n");
    }
    else  
    {
#        SAS::message($SAS::AppMsg, $SAS::SparseMsg, $message);



    }

}
#**************************************
# Subroutine HighlightedMessage
# Outputs lines of text
#**************************************
sub HighlightedMessage($)
{ 
    my ($character, @Messages) = @_;  
   
    my $noLines = @Messages;
    my $maxLength = length($Messages[0]);
    my $i = 0;
    if ($noLines > 1)
    {
        for ($i = 1; $i < $noLines; $i++)
        {
            if (length($Messages[$i]) > $maxLength)
	    {
                $maxLength = length($Messages[$i]);
            }
        } 
    }


    
    my $text = "";
   
    for ($i = 0; $i < $maxLength; $i++)
    { 
        $text = $text . $character;
    }
    &Message("\n$text");
    
    for ($i = 0; $i < $noLines; $i++)
    { 
       
        &Message($Messages[$i]);
    }

    &Message($text);
   

}

#***************************************
# CreateLogFile
#**************************************
sub CreateLogFile
{
   
    $redirect = 5;
   
    open (STDOUT, ">omichain_log"); # and die("Couldn't open file");;
 
    
 
    open (STDERR, ">&STDOUT");
 
 
    select (STDERR);
  
 
}
#***********************************
# Subroutine GetCurrentDirectory
#**********************************
sub GetCurrentDirectory
{
    $command = "pwd";
    $curdir = `$command`;

    $leng = length($curdir);
    substr($curdir, $leng - 1) = "";    
    return $curdir;
}


#******************************
# Subroutine GetODFDirectory
# Retrieves the directory-path
# of the ODF data files.
#******************************
sub GetODFDirectory($)
{
    my ($file) = @_;  
   
    open (SASFILE, $file); #  || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 
 

    $line = <SASFILE>;
        

    while ($line = <SASFILE>)
    {
        
        if ($line =~ "PATH")
        {        

            $length = length($line);
            $directory = substr($line, 5, $length - 6); 
            if ($directory eq "." or $directory eq "./")
            {
                $length = length($SASFile);
                $directory = substr($SASFile, 0, $length - 32);
            }
            close SASFILE;
            return $directory; 
             
        }
    }
    close SASFILE;
#    SAS::fatal("OMICHAIN", "Unable to find the ODF directory path");

}

#************************************
# Sub GetOmmodmapParameters
#************************************
sub GetOmmodmapParameters
{
#    $ommodmap_nsig = intParameter("ommodmapnsig");
#    $ommodmap_nbox = intParameter("ommodmapnbox");
}
#************************************
# Sub GetOmdetectParameters
#************************************
sub GetOmdetectParameters
{
#    $omdetect_nsigma = realParameter("omdetectnsigma");
#    $omdetect_minsignificance = realParameter("omdetectminsignificance");
 

#   $omdetect_detectextended = booleanParameter("omdetectdetectextended");

}
#************************************
# Sub GetOmsrclistcombParameters
#************************************
sub GetOmsrclistcombParameters
{
#    $omsrclistcomb_nsigma = realParameter("omsrclistcombnsigma");
#    $omsrclistcomb_usecat = &GetBooleanParameter("omsrclistcombusecat");
  
}
#************************************
# Sub GetOmmosaicParameters
#************************************
sub GetOmmosaicParameters
{
#    $ommosaic_exposure = realParameter("ommosaicexposure");
#    $ommosaic_exposuremap = &GetBooleanParameter("ommosaicexposuremap");
    
}

#************************************
# Sub GetOmattParameters
#************************************
sub GetOmattParameters
{

#    $omatt_usecat = &GetBooleanParameter("omattusecat");
  
#    $omatt_rotateimage =  &GetBooleanParameter("omattrotateimage");

 
  


}
#**************************************
# GetBooleanParameter
#**************************************
sub GetBooleanParameter($)
{
    my ($name) = @_;
    my ($parameter) = booleanParameter($name);
   
    my ($converted) = "F";
    if ($parameter eq "1" or $parameter eq "yes" or $parameter eq "YES")
    {
        $converted = "T";
    }
  
    return $converted;
}
#**************************************
# Sub GetFilterList
# Returns a list of the OM filters
# in the SAS summary file
#**************************************
sub GetFilterList
{

    my $SASFILE = "SASFILE";


    open ($SASFILE, $SASFile); #  || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 
 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************
   
   
    $line = <$SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {

        $line = <$SASFILE>;

                                  
    } 

    @FilterList = (); 

    while ($line = <$SASFILE>)

    {
        
        if ($line =~ "/ Exposure Identifier")
        {                 


           $filterName = &GetNextSASFilter($SASFILE);
	  
           if ($filterName eq "")
           {
               last;
           }       
           
             
            push(@ExposureFilterList, $filterName);
            #*********************************************
            # Add the filter to the list
            #*********************************************
            $numFilters = @FilterList;
            $exists = 0;
       
        
	    $i = -1;
            while (++$i < $numFilters and $exists == 0) 
            {
	   
                if ($filterName eq $FilterList[$i])
                {
                    print("Line 1768 $filterName $FilterList[$i]\n");
                    $exists = 1;


                }

            }
	    
            if ($exists == 0)  # Found a different filter
            {
 
                if (&FilterOkay($filterName) == 0)
                {
		    print("Line 1781 pushing $filterName onto FilterList\n");
                    push(@FilterList, $filterName);
                }
	    }
        }
    }

    close $SASFILE;
    Display("Filter list", @FilterList);
    return @FilterList;
    
}
#************************************************
# Sub GetFilterFileList
# Returns a list of files for a
# given filter which contain
# the given expression
# Usage :: @List = @GetFilterList("V", "*IMI*");
#************************************************
sub GetFilterFileList($)
{
    my ($filter, $string) = @_;  
    my($i, $j) = 0;
    
    open (SASFILE, $SASFile); #  || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFILE"); 
 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************

 
    $line = <SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {
        $line = <SASFILE>;
                                  
    } 
    @List =();           # Initialise the file list    
 
    while ($line = <SASFILE>)
    {
        
        if ($line =~ "FILTER =")
        {
         
            $pos = index($line, "//");

            $filterName = substr($line, 9, $pos - 10); # Get rid of the characters after the end of the filter
	    
            if ($filterName eq "White")
            {
                $filterName = "WHITE";
	    }
            if ($filterName eq "Grism1")
            {
                $filterName = "GRISM1";
	    }

            if ($filterName eq "Grism2")
            {
                $filterName = "GRISM2";
	      
	    }
             
            if ($filterName eq "Blocked")
            {
                $filterName = "BLOCKED";
	
            }
     




            if ($filterName eq $filter)
            {  
                until ($line =~ "/ Exposure Identifier")
               
                {
	            $line = <SASFILE>;
           
                } 
                 
                $exposureNo = substr($line, 0, 3);
                push(@List, $exposureNo);         
                     
            }  
        }
    }
   
    close SASFILE;
    #******************************************************
    # Select all the files containing the specified string    
    #******************************************************
    @FileList = <$inp_directory/$string>;        

      
    @FilterFileList =();           # Initialise the file list    
 
    for ($i = 0; $i < @FileList; $i++)
    { 
       
        for ($j = 0; $j < @List; $j++)
        {
            if ($FileList[$i] =~ $List[$j])
            {
                push(@FilterFileList, $FileList[$i]);         
            } 
        }
    }
    return @FilterFileList;  
    
}
#*****************************
# Subroutine Display
# Displays the given list
#****************************
sub Display($)
{
    my ($title, @List) = @_;
    my ($i) = 0;
    print("DISPLAYING LIST $title \n"); 
 
  
    for ($i = 0; $i < @List; $i++)
    {
        print("$i) $List[$i] \n");
    }
}
#*******************************************
# GetNextSASFilter
#******************************************
sub GetNextSASFilter($)
{

    my ($SASFILE) = @_;
   
    

    my $line = <$SASFILE>;
 
    
    until($line =~ "FILTER =" or $line eq "")
    {
        $line = <$SASFILE>;
       
    } 
    if ($line eq "")
    {
      
        return $line;

    }
            
    my $pos = index($line, "//");

    my $filterName = substr($line, 9, $pos - 10); # Get rid of the characters after the end of the filter

    if ($filterName eq "White")
    {
        $filterName = "WHITE";
    }


    if ($filterName eq "Grism1")
    {
        $filterName = "GRISM1";
    }

    if ($filterName eq "Grism2")
    {
        $filterName = "GRISM2";
	
    }
             
    if ($filterName eq "Blocked")
    {
        $filterName = "BLOCKED";
	
    }
             
   
    return $filterName;
}
#************************************
# GetExposureNoListForFilter      
# Returns a list of the exposures
# made for the given filter
#************************************
sub GetExposureNoListForFilter($)
{
    my ($filter) = @_;  
    my $i = 0;
    my(@ExposureNoFilterList) = ();           # Initialise the file list    
 
    for ($i = 0; $i < @ExposureNoList; $i++)
    { 
  
        if ($ExposureFilterList[$i] eq $filter)
        {
            push(@ExposureNoFilterList, $ExposureNoList[$i]);         
        } 
        
    }

    return @ExposureNoFilterList;  
}    
#*******************************************
# Sub FindExposureType
# Determines if the given exposure number
# is for a RUDI5 sequence, or ENG2
#*******************************************
sub FindExposureType($)
{
    my ($exposureNo) = @_;  
 
    my(@ImageFiles) = <$inp_directory/*IMI.FIT*>;           


    my ($j, $num);
    $numFiles = @ImageFiles;
  
    @ExposureImageList = ();           # Initialise the file list    
    
 


    for ($j = 0; $j < $numFiles; $j++)
    {
               
        $fileName = $ImageFiles[$j];       
        $pos = index($fileName, "IMI") - 24;

        $string = substr($fileName, $pos, 32);    # Get rid of the directory name

        $text = substr($string, 19, 3);           # The exposure number
    
        if ($text eq $exposureNo)
        {
            push(@ExposureImageList, $fileName);         
          
        }            
              
                  
    }


    $numIMIFiles = @ExposureImageList;
      
    $num = 0;
    
    
    
    if ($numIMIFiles > 0)
    {

        for ($j = 0; $j < $numIMIFiles; $j++)
        {
        
            $fileName = $ExposureImageList[$j];       
            $pos = index($fileName, "IMI") - 24;
            $name = substr($fileName, $pos, 32);    # Get rid of the directory name               
            $pattern = "0" .  $j . "IMI";  
            $pos = index($name, $pattern);
            if ($pos > 0) 
            {
                $num++;
            }
        }
        if ($num == 4)
        {
            $type = "ENG2";
            $ExposureType{$exposureNo} = "ENG2";
	    $ExposureExtension{$exposureNo} = "*ENG2.FIT*";
	
        }
        else
        {
            $type = "RUDI5";
            $ExposureType{$exposureNo} = "RUDI5";
	    $ExposureExtension{$exposureNo} = "*IMI.FIT*";
	
        }
    }
    else
    {


        my(@ImageFiles) = <$inp_directory/*E4I.FIT*>;           
        $numFiles = @ImageFiles;
   

        for ($j = 0; $j < $numFiles; $j++)
        {
               
            $fileName = $ImageFiles[$j];       
            $pos = index($fileName, "E4I") - 24;

            $string = substr($fileName, $pos, 32);    # Get rid of the directory name

            $text = substr($string, 19, 3);           # The exposure number
    
            if ($text eq $exposureNo)
            {
                push(@ExposureImageList, $fileName);         
          
            }            
              
                  
        }


        $numFiles = @ExposureImageList;
      
  
        if ($numFiles == 0 )
        {
	   
            $type = "MISSING";
        }
        else
        {
	    $type = "ENG4";
            $ExposureType{$exposureNo} = "ENG4";
            $ExposureExtension{$exposureNo} = "*E4I.FIT*";
	
        }

    }
    return $type;
    
}
#******************************************************
# Sub MosaicImage
# Mosaics the given list of
# images
#*****************************************************
sub MosaicImages($)
{

print("Mosaic images omatt_rotateimage=$omatt_rotateimage\n");
    if ($omatt_rotateimage eq F) 
    {

        return;
    }
    my($exposureType, $filter, @ImageList) = @_;
    $text = "ommosaic- mosaicing images for filter $filter";
    &HighlightedMessage("*", $text);
    
     

    if ($exposureType eq "RUDI5")
    { 
        $string = "RSIMAG";
    }
    elsif($exposureType eq "ENG2")
    {
        $string = "LSIMAG";
    }
    elsif($exposureType eq "ENG4")
    {
        $string = "HSIMAG";
    }
    elsif($exposureType eq "MISSING")
    {
        return;
    }
    else

    {
        die ("UNIDENTIFIED EXPOSURE TYPE ($exposureType)");
    }
            
    my $filterCode = &GetFilterCode($filter);

    my $position = index($ImageList[0], "OMS") - 11;
    
   
    my $outputFileName = substr($ImageList[0], $position, 14) . "000$string" . $filterCode . ".FIT";
     

   
    $arg_list = "imagesets=\"" . join(" ", @ImageList) . "\" mosaicedset=$outputFileName"; #. "exposuremap=$ommosaic_exposuremap " . "exposure=$ommosaic_exposure";
#    print("arg-list = $arg_list \n");

    &Message("ommosaic $arg_list");
    system("ommosaic $arg_list"); #  && SAS::fatal("Task Failure", "ommosaic");

}
#************************************************
# Sub GetFilterFileList
# Returns a list of files for a
# given filter which contain
# the given expression
# Usage :: @List = @GetFilterList("V", "*IMI*");
#************************************************
sub GetExposureFileList($)
{

    my ($directory, $exposureNo, $string) = @_;  
    my $i = 0;

   


    #******************************************************
    # Select all the files containing the specified string    
    #******************************************************
    my(@FileList) = <$directory/$string>;        
    
    my @ExposureFileList =();           # Initialise the file list    
 
    for ($i = 0; $i < @FileList; $i++)
    { 
        my $position = length($FileList[$i]) - 12; 
        if (substr($FileList[$i], $position, 3) =~ $exposureNo)
  
        {
	    
            push(@ExposureFileList, $FileList[$i]);         
        } 
    }
    
    return @ExposureFileList;  
    
}
#***************************************
# Subroutine CheckRudi5Images
#***************************************
sub CheckRudi5Images($)
{
    my ($exposureNo, @FileList) = @_; 

    my $number = @FileList;

    if ($number ne 2)
    {

        $text = "Warning! - Missing image file";
        &HighlightedMessage("X", $text);
    }
}
#******************************************
# Subroutine CombineEng2Images
# Combines the 4 separate ENG2 image files
# into a single image
#******************************************    
sub CombineEng2Images($)
{
 
    my ($observationNo, $exposureSchedule, $exposureNo) = @_;
            
    my($fileName) = $ExposureImageList[0];       
    my($pos) = index($fileName, "IMI") - 24;
    my($name) = substr($fileName, $pos, 32);    # Get rid of the directory name                          

    #*****************************************************************
    # The new image file is temporarily stored in the input directory 
    #*****************************************************************  

    $outputFileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "00E2I.FIT";
  
  
    

    &HighlightedMessage("*", "omcomb");
    $arg_list = "imagesets=\"" . join(" ", @ExposureImageList) . "\"" . " outset=$outputFileName";

    &Message("omcomb $arg_list");
              
    system("omcomb $arg_list"); #  && SAS::fatal("Task Failure", "omcomb");
    #***********************************************
    # Check to see if omcomb ran successfully
    #***********************************************
    $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
   
    if ($exit_value == 1)  
    { 
       

	$ExposureType{$exposureNo} = "RUDI5";
	$ExposureExtension{$exposureNo} = "*IMI.FIT*";
	                
     
        &Message("The test in omcomb have failed - This exposure ($exposureNo) will be treated as normal imaging data");
        
    }      
    else
    {

                          
        $ExposureType{$exposureNo} = "ENG2";
	$ExposureExtension{$exposureNo} = "*E2I.FIT*";
		
        push(@E2IMIList, $outputFileName);   # A list of the combined image files           
        	
    }  
}
#************************************
# Subroutine FilterOkay
#************************************
sub FilterOkay($)
{
    my ($filter) = @_;
    my $i;
    my (@BadFilter) = ("BLOCKED", "Blocked");
    
  
    for ($i = 0; $i < @BadFilter; $i++)
    {
	
        if ($filter eq $BadFilter[$i])
        {
            $message = "Warning! - Cannot process filter $filter. The image files for this filter will not be processed";
            &HighlightedMessage("X", $message);
         
            return 1;
        }
    }
    
    return 0;
}
#***************************
# Subroutine GetFileList
#***************************
sub GetFileList($)
{
    my (@Pattern) = @_;
    my @Filelist = ();
    my ($i, $j);

    
    for ($i = 0; $i < @Pattern; $i++)
    {
	
        my (@List) = <$Pattern[$i]>;
        for ($j = 0; $j < @List; $j++)
        {
            push(@FileList, $List[$j]);
        }
    }
    &Display("test", @FileList);
    return @FileList;
}         
#*****************************************
# Checks to see if the given file exists
# Returns 1 if it does, 0 otherwise.
#
#*****************************************
sub FileExists($)
{
    my ($file) = @_;
    if (-e $file)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}
#***************************
# GetStringParameterList
# Gets the list of strings
# in the command line for a
# specified parameter name
#***************************
sub GetStringParameterList($)
{
    my($parameter) = @_;
    my $num = 0;
    my $i = 0;
#    $num = parameterCount($parameter);
    my (@ParameterList) = ();
   
    if ($num == 0)
    {
        return @ParameterList;
    }
  

   for ($i = 0; $i < $num; $i++)
   {
#        $ParameterList[$i] = stringParameter($parameter, $i);

   }

       


    return @ParameterList; 
   
}

#************************************************************
# Sub CheckUserFilterList
# Checks that the user-supplied list of filters exist in 
# the SAS summary file and returns a new list (if necessary)
#************************************************************  
sub CheckUserFilterList($)
{
    my(@UserFilterList) = @_;
    my(@List) = &GetFilterList;  # List of OM filters in summary file
      
    my ($i, $j) = 0;
    my @ModifiedList = ();

    for ($i = 0; $i < @UserFilterList; $i++)
    {
        my $found = 0;
        my $j = 0;
        while ($j lt @List and $found == 0)
        {
            if ($List[$j] eq $UserFilterList[$i])
            {  
                push(@ModifiedList,  $UserFilterList[$i]);
                $found = 1;
            }
            $j++;
        }
        if ($found == 0)
        {
            $message = "Warning! - Filter $UserFilterList[$i] does not exist in SAS Summary file- ignoring filter";
            &HighlightedMessage("X", $message); 
        }
    }
    if (@ModifiedList == 0)
    {
        @OMFilterList = &GetFilterList; # The list of OM filters from the SAS summary file
     
        $message = "Error! - No valid filters to process (Valid filters are ". join(", ", @OMFilterList) .")- Aborting";
    
        &HighlightedMessage("X", $message);
       
        die;
    }
    return @ModifiedList;
}
#*******************************
# Sub CheckUserExposureList
# Checks that the command-line 
# list of exposures is valid
# (ie the filters are present)
# and get a list of the filters
# for the exposure list
#*******************************
sub CheckUserExposureList($)
{
    my (@UserExposureList) = @_;         # The list of exposures in the parameter list
    my (@OMFilterList) = &GetFilterList; # The list of OM filters from the SAS summary file
    my (@FilterList) = ();
    my (@ExposureList) = ();
    my $i = 0;

   

    #**************************************************
    # Loop through the user-supplied list of exposures
    # and produce a filter list
    #**************************************************


    for ($i = 0; $i < @UserExposureList; $i++)
    {
        my $found = 0;
        my $j = 0;

        #******************************************************************
        # Loop through the filter list obtained from the SAS summary file
        #******************************************************************
        while ($j < @OMFilterList and $found == 0)
        {

  

            #****************************************************
            # Get a list of the exposures for this filter
            #****************************************************
            my (@ExposureNoList) = &GetExposureNoListForFilter($OMFilterList[$j]);
            my $k = 0;
            while ($k < @ExposureNoList and $found == 0)
            {
                push(@ExposureList, $ExposureNoList[$k]); 
                if ($ExposureNoList[$k] eq $UserExposureList[$i])
                {
                    $filterExists = 0;

#                    for ($l = 0; $l < @FilterList; $l++)
		    $l = 0;
                    while($l++ < @FilterList and $filterExists == 0)
                    {
                        if ($FilterList[$l] eq  $OMFilterList[$j])
                        {
                            $filterExists = 1;
#                            break;
                        }
                    }
                    if ($filterExists == 0) 
                    {
                        push(@FilterList, $OMFilterList[$j]);
                    } 
                    $found = 1;
                  
                }
                $k++;
            }
            $j++;
	}
        if ($found == 0) 
        {
            &HighlightedMessage("X", "Warning! - Invalid exposure no ($UserExposureList[$i])");
        }  
           
    }
  
    if (@FilterList == 0)
    {

        my @Message = ();
        push (@message, "Error!- Invalid list of exposures- aborted");
        for ($i = 0; $i < @OMFilterList; $i++ )
        {
            my (@ExposureNoList) = &GetExposureNoListForFilter($OMFilterList[$i]);
            $text = "Valid exposures for filter $OMFilterList[$i] are "; 
      
            $text = $text . join(", ", @ExposureNoList);   
            push(@message, $text);
	}
        &HighlightedMessage("X", @message);
    }
 
  
    return @FilterList;
}                 








